@@ -1,23 +1,20 @@
 import base64
 import copy
-import glob
 import importlib
 import importlib.util as util
 import inspect
 import json
 import os
-import site
-import socket
-import sqlite3
-import sys
-import winreg
+if os.name == 'nt':
+    import winreg
+else:
+    import site
 import xml.etree.ElementTree as ElTree
 import zipfile
 import zlib
 from datetime import datetime, timedelta
 from inspect import signature
-from shutil import copyfile
-from typing import List, Any
+from sqlite3 import connect
 from urllib import parse
 
 import xmltodict
@@ -40,7 +37,7 @@ import xmltodict
 
 class WorkflowEngine:
 
-    def __init__(self, input_parameter: Any = None, pythonpath: str = "", installation_directory: str = ""):
+    def __init__(self, input_parameter: any = None, pythonpath: str = "", installation_directory: str = ""):
         """
         Class for automating DrawIO diagrams
         :param input_parameter: An object holding arguments to be passed as input to the WorkflowEngine.  in a flow, use get_input_parameter to retrieve the value.
@@ -57,72 +54,77 @@ class WorkflowEngine:
             if os.name == 'nt':
                 pythonpath = self.get_python_path()
             else:
-                with open('settings') as json_file:
-                    data = json.load(json_file)
-                    pythonpath = data["pythonpath"]
-                    settings.update({'pythonpath': pythonpath})
+                sett = '/etc/BPMN_RPA_settings'
+                if not os.path.exists(sett):
+                    file = open(sett, "w")
+                    file.write("{\"pythonpath\": \"\", \"dbpath\":\"\"}")
+                    file.close()
+                json_file = open(sett, "r")
+                data = json.load(json_file)
+                json_file.close()
+                pythonpath = data["pythonpath"]
         if len(installation_directory) != 0:
-            if os.name == 'nt':
-                self.set_db_path(installation_directory)
-            else:
-                settings.update({'dbpath': installation_directory})
+            self.set_db_path(installation_directory)
             db_folder = installation_directory
         else:
+            db_folder = self.get_db_path()
+        if db_folder is None or len(db_folder) == 0:
             if os.name == 'nt':
-                db_folder = self.get_db_path()
+                message = "\nYour installation directory is unknown. Please enter the path of your installation directory: "
             else:
-                with open('settings') as json_file:
-                    data = json.load(json_file)
-                    db_folder = data["dbpath"]
-                    settings.update({'dbpath': db_folder})
-        if db_folder is None:
-            installdir = input(
-                "\nYour installation directory is unknown. Please enter the path of your installation directory: ")
-            if not str(installdir).endswith("\\"):
-                installdir += "\\"
+                message = "\nYour installation folder is unknown. Please enter the path of your installation folder: "
+            installdir = input(message)
+            if os.name == 'nt':
+                if not str(installdir).endswith("\\"):
+                    installdir += "\\"
+            else:
+                if not str(installdir).endswith("/"):
+                    installdir += "/"
             if not os.path.exists(installdir):
                 os.mkdir(installdir)
             if len(installdir) == 0:
                 return
             else:
-                if os.name == 'nt':
-                    self.set_db_path(installdir)
-                else:
-                    settings.update({'dbpath': installdir})
+                self.set_db_path(installdir)        
                 db_folder = installdir
-        if pythonpath is None:
-            pythonpath = input(
-                "\nThe path to your Python.exe file is unknown. Please enter the path to your Python.exe file: ")
+        if pythonpath is None or len(pythonpath) == 0:
+            if os.name == 'nt':
+                message = "\nThe path to your Python.exe file is unknown. Please enter the path to your Python.exe file: "
+            else:
+                message = "\nEnter the full path of your Python installation:"
+            pythonpath = input(message)
             if not os.path.exists(pythonpath):
-                if os.name == 'nt':
-                    self.set_python_path(pythonpath)
-                else:
-                    settings.update({'pythonpath': pythonpath})
+                self.set_python_path(pythonpath)
             if len(pythonpath) == 0:
                 return
             else:
-                if os.name == 'nt':
-                    self.set_python_path(pythonpath)
-                else:
-                    settings.update({'pythonpath': pythonpath})
+                self.set_python_path(pythonpath)
         if os.name != 'nt':
-            with open('settings', 'w') as outfile:
+            settings.update({'dbpath': db_folder})
+            settings.update({'pythonpath': pythonpath})
+            with open('/etc/BPMN_RPA_settings', 'w') as outfile:
                 json.dump(settings, outfile)
         self.input_parameter = input_parameter
         self.pythonPath = pythonpath
+        if os.name == 'nt':
+            self.packages_folder = "\\".join(pythonpath.split('\\')[0:-1]) + "\\Lib\\site-packages"
+        else:
+            self.packages_folder = site.getsitepackages()[0]
         self.db = SQL(db_folder)
         self.db.orchestrator()  # Run the orchestrator database
         self.id = -1  # Holds the ID for our flow
-        self.error = False  # Indicator if the flow has any errors in its execution
+        self.error = None  # Indicator if the flow has any errors in its execution
         self.step_name = None
         self.flowname = None
         self.flowpath = None
         self.loopvariables = []
         self.previous_step = None
-        self.step_nr = None
+        self.step_nr = 0
+        self.step_input = None
+        self.runlog = ""
         self.variables = {}  # Dictionary to hold WorkflowEngine variables
 
-    def get_input_parameter(self, as_dictionary: bool = False) -> Any:
+    def get_input_parameter(self, as_dictionary: bool = False) -> any:
         """
         Returns the input parameter that was given when creating an instance of the WorkflowEngine
         :param as_dictionary: Optional. Indicator whether to treat the given input as a dictionary object (string to dict).
@@ -130,11 +132,13 @@ class WorkflowEngine:
         """
         if as_dictionary:
             if not isinstance(self.input_parameter, dict):
+                if isinstance(self.input_parameter, str):
+                    self.input_parameter = self.input_parameter.replace("true", "True").replace("false", "False")
                 self.input_parameter = eval(self.input_parameter)
         self.print_log(f"Got input parameter {str(self.input_parameter)}")
         return self.input_parameter
 
-    def open(self, filepath: str, as_xml: bool = False) -> Any:
+    def open(self, filepath: str, as_xml: bool = False) -> any:
         """
         Open a DrawIO document
         :param filepath: The full path (including extension) of the diagram file
@@ -147,15 +151,16 @@ class WorkflowEngine:
         xml_file = open(filepath, "r")
         if not filepath.__contains__(".vsdx"):
             if filepath.__contains__(".flw"):
-                self.flowname = filepath.split("\\")[-1].lower().replace(".flw", "")
-                with open(filepath, encoding="utf8", errors='ignore') as f:
-                    content = f.read()
-                decoded = base64.b64decode(content).decode("utf-8", errors='ignore')
-                idx = decoded.index("}]")
-                dict_list = json.loads(decoded[0:idx + 2])
+                self.flowname = filepath.split("\\")[-1].replace(".flw", "")
+                with open(filepath, "rb") as binary_file:
+                    # Read the whole file at once
+                    content = binary_file.read()
+                str_content = content[24:-1].decode("utf-8", errors='ignore')
+                decoded = base64.b64decode(str_content).decode("utf-8", errors='ignore')
+                dict_list = json.loads(decoded)
                 return dict_list
             else:
-                self.flowname = filepath.split("\\")[-1].lower().replace(".xml", "")
+                self.flowname = filepath.split("\\")[-1].replace(".xml", "")
                 xml_root = ElTree.fromstring(xml_file.read())
                 raw_text = xml_root[0].text
                 base64_decode = base64.b64decode(raw_text)
@@ -168,7 +173,7 @@ class WorkflowEngine:
         else:
             # It is a MsVisio file!
             visio = Visio()
-            self.flowname = filepath.split("\\")[-1].lower().replace(".vsdx", "")
+            self.flowname = filepath.split("\\")[-1].replace(".vsdx", "")
             visio.open_vsdx_file(filepath)
             return visio
 
@@ -178,46 +183,67 @@ class WorkflowEngine:
         Write the orchestrator database path to the registry.
         :param value: The path of the orchestrator database that has to be written to the registry
         """
-        try:
-            reg_path = r"SOFTWARE\BPMN_RPA"
-            winreg.CreateKey(winreg.HKEY_CURRENT_USER, reg_path)
-            registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0,
-                                          winreg.KEY_WRITE)
-            winreg.SetValueEx(registry_key, "dbPath", 0, winreg.REG_SZ, value)
-            winreg.CloseKey(registry_key)
-            return True
-        except WindowsError:
-            return False
+        if os.name == 'nt':
+            try:
+                reg_path = r"SOFTWARE\BPMN_RPA"
+                winreg.CreateKey(winreg.HKEY_CURRENT_USER, reg_path)
+                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0,
+                                              winreg.KEY_WRITE)
+                winreg.SetValueEx(registry_key, "dbPath", 0, winreg.REG_SZ, value)
+                winreg.CloseKey(registry_key)
+                return True
+            except WindowsError:
+                return False
+        else:
+            json_file = open('/etc/BPMN_RPA_settings', 'r')
+            data = json.load(json_file)
+            json_file.close()
+            data["dbpath"] = value
+            json_file = open('/etc/BPMN_RPA_settings', 'w')
+            json.dump(data, json_file)
+            json_file.close()
 
     @staticmethod
-    def get_db_path() -> Any:
+    def get_db_path() -> any:
         """
         Get the path to the orchestrator database
         :return: The path to the orchestrator database
         """
-        try:
-            reg_path = r"SOFTWARE\BPMN_RPA"
-            registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_READ)
-            value, regtype = winreg.QueryValueEx(registry_key, 'dbPath')
-            winreg.CloseKey(registry_key)
-            return value
-        except WindowsError:
-            return None
+        if os.name == 'nt':
+            try:
+                reg_path = r"SOFTWARE\BPMN_RPA"
+                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_READ)
+                value, regtype = winreg.QueryValueEx(registry_key, 'dbPath')
+                winreg.CloseKey(registry_key)
+                return value
+            except WindowsError:
+                return None
+        else:
+            json_file = open('/etc/BPMN_RPA_settings', 'r')
+            data = json.load(json_file)
+            json_file.close()
+            return data["dbpath"]
 
     @staticmethod
-    def get_python_path() -> Any:
+    def get_python_path() -> any:
         """
         Get the path to the Python.exe file
         :return: The path to the Python.exe file
         """
-        try:
-            reg_path = r"SOFTWARE\BPMN_RPA"
-            registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_READ)
-            value, regtype = winreg.QueryValueEx(registry_key, 'PythonPath')
-            winreg.CloseKey(registry_key)
-            return value
-        except WindowsError:
-            return None
+        if os.name == 'nt':
+            try:
+                reg_path = r"SOFTWARE\BPMN_RPA"
+                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0, winreg.KEY_READ)
+                value, regtype = winreg.QueryValueEx(registry_key, 'PythonPath')
+                winreg.CloseKey(registry_key)
+                return value
+            except WindowsError:
+                return None
+        else:
+            json_file = open('/etc/BPMN_RPA_settings', 'r')
+            data = json.load(json_file)
+            json_file.close()
+            return data["pythonpath"]
 
     @staticmethod
     def set_python_path(value: str):
@@ -225,18 +251,27 @@ class WorkflowEngine:
         Write the oPython path to the registry.
         :param value: The path of the Python.exe file that has to be written to the registry
         """
-        try:
-            reg_path = r"SOFTWARE\BPMN_RPA"
-            winreg.CreateKey(winreg.HKEY_CURRENT_USER, reg_path)
-            registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0,
-                                          winreg.KEY_WRITE)
-            winreg.SetValueEx(registry_key, "PythonPath", 0, winreg.REG_SZ, value)
-            winreg.CloseKey(registry_key)
-            return True
-        except WindowsError:
-            return False
+        if os.name == 'nt':
+            try:
+                reg_path = r"SOFTWARE\BPMN_RPA"
+                winreg.CreateKey(winreg.HKEY_CURRENT_USER, reg_path)
+                registry_key = winreg.OpenKey(winreg.HKEY_CURRENT_USER, reg_path, 0,
+                                              winreg.KEY_WRITE)
+                winreg.SetValueEx(registry_key, "PythonPath", 0, winreg.REG_SZ, value)
+                winreg.CloseKey(registry_key)
+                return True
+            except WindowsError:
+                return False
+        else:
+            json_file = open('/etc/BPMN_RPA_settings', 'r')
+            data = json.load(json_file)
+            json_file.close()
+            data["pythonpath"] = value
+            json_file = open('/etc/BPMN_RPA_settings', 'w')
+            json.dump(data, json_file)
+            json_file.close()
 
-    def get_flow(self, ordered_dict: Any) -> Any:
+    def get_flow(self, ordered_dict: any) -> any:
         """
         Retrieving the elements of the flow in the Document.
         :param ordered_dict: The document object containing the flow elements.
@@ -301,7 +336,7 @@ class WorkflowEngine:
         retn = shapes + connectors
         return retn
 
-    def get_step_from_shape(self, shape: Any) -> Any:
+    def get_step_from_shape(self, shape: any) -> any:
         """
         Build a Step-object from the Shape-object
         :param shape: The Shape-object
@@ -331,7 +366,7 @@ class WorkflowEngine:
         return retn
 
     @staticmethod
-    def get_variables_from_text(text: str) -> Any:
+    def get_variables_from_text(text: str) -> any:
         """
         Get variable names (like '%variable%') from text.
         :param text: The text to get the variables from
@@ -358,7 +393,7 @@ class WorkflowEngine:
             retn = None
         return retn
 
-    def get_parameters_from_shapevalues(self, step: Any, input_signature: Any) -> Any:
+    def get_parameters_from_shapevalues(self, step: any, input_signature: any) -> any:
         """
         If input values are provided in the Shape values, then create a mapping
         :param step: The step to use the Shape values of to create the mapping
@@ -489,7 +524,7 @@ class WorkflowEngine:
                                         else:
                                             if str(replace_value).__contains__("Message(mime_content="):
                                                 if attr is None:
-                                                    attr=""
+                                                    attr = ""
                                                 if isinstance(replace_value, list):
                                                     if tv.__contains__("."):
                                                         attr = str(lst[0].split(".")[1]).replace(".", "")
@@ -507,7 +542,7 @@ class WorkflowEngine:
                                             else:
                                                 replace_value = self.get_attribute_value(lst[0], replace_value)
                                                 if val is not None and replace_value is not None:
-                                                    val = val.replace(tv, replace_value)
+                                                    val = val.replace(tv, str(replace_value))
                                                 else:
                                                     val = replace_value
                                 else:
@@ -563,7 +598,7 @@ class WorkflowEngine:
             return mapping
 
     @staticmethod
-    def get_attribute_value(lst: str, replace_value: Any) -> Any:
+    def get_attribute_value(lst: str, replace_value: any) -> any:
         """
         Get an attribute value from a replace value (object)
         :param lst: The attribute list
@@ -590,7 +625,7 @@ class WorkflowEngine:
         return val
 
     @staticmethod
-    def step_has_direct_variables(step: Any) -> bool:
+    def step_has_direct_variables(step: any) -> bool:
         """
         Check if a step uses any variables as input for any of the Shapevalue fields
         :param step: The step to check
@@ -604,41 +639,46 @@ class WorkflowEngine:
         else:
             return False
 
-    def run_flow(self, steps: Any):
+    def run_flow(self, steps: any, step_by_step: bool = False):
         """
         Execute a Flow.
-        :params steps: The steps that must be executed in the flow
+        :param steps: The steps that must be executed in the flow
+        :param step_by_step: Optional. Indicator if this function only performes one step and the looping of steps is done outside this function.
         """
+        step = None
+        output_previous_step = None
+        if not isinstance(steps, list):
+            steps = [steps]
+            step = steps[0]
         db_path = self.get_db_path()
-        if db_path == "\\":
-            self.error = True
-            raise Exception('Your installation directory is unknown.')
-        # Save the flow if not already saved
-        # if not os.path.exists(f'{self.db}\\Registered Flows\\{self.flowname}.xml') and not os.path.exists(
-            #   self.flowpath):
-            # Move the file to the registered directory if not exists
-            # copyfile(self.flowpath, f'{db_path}\\Registered Flows\\{self.flowname}.xml')
-        # if not self.flowpath.lower().__contains__(".vsdx"):
-        #     self.flowpath = f'{db_path}\\Registered Flows\\{self.flowname}.xml'
-        if not str(self.flowpath).__contains__("\\"):
-            self.flowpath = os.getcwd() + f"\\{self.flowpath}"
+        if os.name == 'nt':
+            if db_path == "\\":
+                self.error = True
+                raise Exception('Your installation directory is unknown.')
+            if not str(self.flowpath).__contains__("\\"):
+                self.flowpath = os.getcwd() + f"\\{self.flowpath}"
+        else:
+            if db_path == "/":
+                self.error = True
+                raise Exception('Your installation directory is unknown.')
+            if not str(self.flowpath).__contains__("/"):
+                self.flowpath = os.getcwd() + f"/{self.flowpath}"
         sql = f"SELECT id FROM Flows WHERE name ='{self.flowname}' AND location='{self.flowpath}'"
         flow_id = self.db.run_sql(sql=sql, tablename="Flows")
         if flow_id is None:
             sql = f"INSERT INTO Flows (name, location) VALUES ('{self.flowname}','{self.flowpath}');"
             flow_id = self.db.run_sql(sql=sql, tablename="Flows")
-        self.previous_step = None
-        output_previous_step = None
-        shape_steps = [x for x in steps if x.type == "shape"]
-        step = [x for x in shape_steps if x.IsStart][0]
-
-        # Log the start in the orchestrator database
-        sql = f"INSERT INTO Runs (name, flow_id, result) VALUES ('{self.flowname}', {flow_id}, 'The flow was aborted.');"
-        self.id = self.db.run_sql(sql=sql, tablename="Runs")
-        print("\n")
-        self.print_log(status="Starting",
-                       result=f"{datetime.today().strftime('%d-%m-%Y')} Starting flow '{self.flowname}'...")
-        self.step_nr = 0
+        if step_by_step is False or self.step_nr == 0:
+            self.previous_step = None
+            shape_steps = [x for x in steps if x.type == "shape"]
+            step = [x for x in shape_steps if x.IsStart][0]
+            # Log the start in the orchestrator database
+            sql = f"INSERT INTO Runs (name, flow_id, result) VALUES ('{self.flowname}', {flow_id}, 'The flow was aborted.');"
+            self.id = self.db.run_sql(sql=sql, tablename="Runs")
+            print("\n")
+            self.print_log(status="Starting",
+                           result=f"{datetime.today().strftime('%d-%m-%Y')} Starting flow '{self.flowname}'...")
+            self.step_nr = 0
         while True:
             try:
                 # to fetch module
@@ -679,11 +719,12 @@ class WorkflowEngine:
                     if method_to_call is None:
                         step_input = None
                         if hasattr(step, "module"):
-                            if not str(step.module).__contains__("\\") and str(step.module).lower().__contains__(".py"):
-                                step.module = f"{site.getsitepackages()[1]}\\BPMN_RPA\\Scripts\\{step.module}"
-                            if not str(step.module).__contains__(":") and str(step.module).__contains__("\\") and str(
+                            if os.name == 'nt':
+                                if not str(step.module).__contains__("\\") and str(step.module).lower().__contains__(".py"):
+                                    step.module = f"{self.packages_folder}\\BPMN_RPA\\Scripts\\{step.module}"
+                                if not str(step.module).__contains__(":") and str(step.module).__contains__("\\") and str(
                                     step.module).__contains__(".py"):
-                                step.module = f"{site.getsitepackages()[1]}\\{step.module}"
+                                    step.module = f"{self.packages_folder}\\{step.module}"
                             if str(step.module).lower().__contains__(".py"):
                                 spec = util.spec_from_file_location(step.module, step.module)
                                 module_object = util.module_from_spec(spec)
@@ -735,6 +776,7 @@ class WorkflowEngine:
                     step_input = self.get_input_from_signature(step, method_to_call)
                 if method_to_call is None and class_object is not None:
                     step_input = self.get_input_from_signature(step, class_object)
+                self.step_input = step_input
 
                 # execute function call and get returned values
                 if step_input is not None and not is_in_loop:
@@ -782,7 +824,6 @@ class WorkflowEngine:
                     this_step = self.loopcounter(step, output_previous_step)
                 if is_in_loop:
                     output_previous_step = [this_step]
-
                 # Update the result
                 if hasattr(step, "classname"):
                     if len(step.classname) == 0:
@@ -804,8 +845,9 @@ class WorkflowEngine:
                         if hasattr(step, "name"):
                             if len(step.name) > 0:
                                 self.print_log(status="Running", result=f"{step.name} executed.")
-            except Exception as e:
-                raise Exception(f"Error: {e}")
+            except Exception as ex:
+                self.set_error(ex)
+                raise Exception(f"Error: {ex}\n{self.error}")
             if step is None:
                 self.end_flow()
                 break
@@ -816,39 +858,66 @@ class WorkflowEngine:
                 if this_step is not None:
                     self.save_output_variable(step, this_step, output_previous_step)
             self.previous_step = copy.deepcopy(step)
-            if self.error:
-                break
+            if step_by_step:
+                return output_previous_step
             step = self.get_next_step(step, steps, output_previous_step)
         if output_previous_step is not None:
             return output_previous_step
 
+    def set_error(self, ex: any):
+        """
+        Set the internal error comming from the try-except
+        :param ex: the exception
+        """
+        trace = []
+        tb = ex.__traceback__
+        while tb is not None:
+            trace.append({
+                "filename": tb.tb_frame.f_code.co_filename,
+                "name": tb.tb_frame.f_code.co_name,
+                "lineno": tb.tb_lineno
+            })
+            tb = tb.tb_next
+            err_ = str({
+                'type': type(ex).__name__,
+                'message': str(ex),
+                'trace': trace
+            })
+        self.error = err_
+
     def print_log(self, result: str, status: str = ""):
         """
         Log progress to the Orchestrator database and print progress on screen
         :param status: Optional. The status of the step
         :param result: The result of the step
         """
-        result = str(result).replace("<br>", " ")
-        result = str(result[0]).capitalize() + result[1:]
-        status = str(status)
-        if not result.endswith("."):
-            result += "."
-        step_time = datetime.now().strftime("%H:%M:%S")
-        if self.step_nr is not None:
-            print(f"{step_time}: Step {self.step_nr} - {result}")
-        else:
-            print(f"{step_time}: {result}")
-            self.step_nr = ""
-        result = result.replace("'", "''").strip()
-        if len(status) > 0:
-            status = f" - {status}"
-        if self.step_name is not None:
-            step_name = self.step_name.replace("'", "''")
-        else:
-            result = "Starting"
-            step_name = "Start"
-        sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', '{step_name}', '{result}', '{self.step_nr}{status}');"
-        self.db.run_sql(sql)
+        try:
+            result = str(result).replace("<br>", " ")
+            result = str(result[0]).capitalize() + result[1:]
+            status = str(status)
+            if not result.endswith("."):
+                result += "."
+            step_time = datetime.now().strftime("%H:%M:%S")
+            if self.step_nr is not None:
+                print(f"{step_time}: Step {self.step_nr} - {result}")
+                self.runlog += f"{step_time}: Step {self.step_nr} - {result}"
+            else:
+                print(f"{step_time}: {result}")
+                self.runlog += f"{step_time}: {result}"
+                self.step_nr = ""
+            result = result.replace("'", "''").strip()
+            if len(status) > 0:
+                status = f" - {status}"
+            if self.step_name is not None:
+                step_name = self.step_name.replace("'", "''")
+            else:
+                result = "Starting"
+                step_name = "Start"
+            sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', '{step_name}', '{result}', '{self.step_nr}{status}');"
+            self.db.run_sql(sql)
+        except Exception as ex:
+            self.set_error(ex)
+            raise Exception(self.error)
 
     def exitcode_not_ok(self):
         """
@@ -870,24 +939,29 @@ class WorkflowEngine:
         """
         # Flow has ended. Log the end in the orchestrator database.
         ok = "The flow has ended."
-        if self.error:
-            ok = "The flow has ended with ERRORS."
-        sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', 'End', 'Ended', '{ok}');"
-        step_time = datetime.now().strftime("%H:%M:%S")
-        finished = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
-        end_result = f"{step_time}: Flow '{self.flowname}': {ok}"
-        print(end_result)
-        self.db.run_sql(sql=sql, tablename="Steps")
-        # Update the result of the flow
-        sql = f"UPDATE Runs SET result= '{ok}', finished='{finished}' where id = {self.id};"
-        self.db.run_sql(sql=sql, tablename="Runs")
+        try:
+            if self.error:
+                ok = "The flow has ended with ERRORS."
+            sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', 'End', 'Ended', '{ok}');"
+            step_time = datetime.now().strftime("%H:%M:%S")
+            finished = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
+            end_result = f"{step_time}: Flow '{self.flowname}': {ok}"
+            print(end_result)
+            self.db.run_sql(sql=sql, tablename="Steps")
+            # Update the result of the flow
+            sql = f"UPDATE Runs SET result= '{ok}', finished='{finished}' where id = {self.id};"
+            self.db.run_sql(sql=sql, tablename="Runs")
+        except Exception as ex:
+            self.set_error(ex)
+            raise Exception(f"Error: {ex}\n{self.error}")
 
-    def get_input_from_signature(self, step: Any, method_to_call: Any) -> Any:
+    def get_input_from_signature(self, step: any, method_to_call: any) -> any:
         sig = None
         try:
             sig = signature(method_to_call)
-        except Exception as e:
-            print(f"Error in getting input from input_signature: {e}")
+        except Exception as ex:
+            self.set_error(ex)
+            print(f"Error in getting input from input_signature: {self.error}")
         if str(sig) != "()":
             step_input = self.get_parameters_from_shapevalues(step=step, input_signature=sig)
             return step_input
@@ -914,7 +988,7 @@ class WorkflowEngine:
                     self.print_log(f"Loopcounter reset for loopvariable '{reset_for_loop_variable}'",
                                    "Running")
 
-    def loopcounter(self, step: Any, output_previous_step: Any) -> Any:
+    def loopcounter(self, step: any, output_previous_step: any) -> any:
         """
         Process steps with a loopcounter
         :param step: The current step object
@@ -972,11 +1046,12 @@ class WorkflowEngine:
                     return loopvar.items[loopvar.counter]
                 else:
                     return output_previous_step
-            except Exception as e:
-                sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', '{step.name}', 'Running', '', 'Error: {e}');"
+            except Exception as ex:
+                self.set_error(ex)
+                sql = f"INSERT INTO Steps (run, name, step, status, result) VALUES ('{self.id}', '{self.flowname}', '{step.name}', 'Running', '', 'Error: {self.error}');"
                 self.db.run_sql(sql=sql, tablename="Steps")
                 self.error = True
-                print(f"Error: {e}")
+                print(f"Error: {self.error}")
                 return output_previous_step
         else:
             return output_previous_step
@@ -1020,8 +1095,6 @@ class WorkflowEngine:
                     {'%__folder_downloads__%': os.path.join(os.path.join(os.environ['USERPROFILE']), 'Downloads')})
             if str(getattr(step, value)).__contains__("%__folder_system__%"):
                 self.variables.update({'%__folder_system__%': os.environ['WINDIR'] + "\\System\\"})
-            if str(getattr(step, value)).__contains__("%__system_name__%"):
-                self.variables.update({'%__system_name__%': socket.getfqdn()})
             if str(getattr(step, value)).__contains__("%__user_name__%"):
                 self.variables.update({'%__user_name__%': os.getenv('username')})
 
@@ -1065,7 +1138,7 @@ class WorkflowEngine:
             retn = False
         return retn
 
-    def get_next_step(self, current_step, steps, output_previous_step: Any) -> Any:
+    def get_next_step(self, current_step, steps, output_previous_step: any) -> any:
         """
         Get the next step in the flow
         :param output_previous_step: The output of the previous step.
@@ -1127,8 +1200,9 @@ class WorkflowEngine:
                     loopvar.counter = loopvar.start
                     loopvar.total_listitems = 0
                     self.loopvariables.append(loopvar)
-                except Exception as e:
-                    print(f"Error: {e}")
+                except Exception as ex:
+                    self.set_error(ex)
+                    print(f"Error: {self.error}")
         return retn
 
     class dynamic_object(object):
@@ -1152,13 +1226,21 @@ class SQL:
         Class for SQLite actions on the Orchestrator database.
         :param dbfolder: The folder of the SQLite orchestrator database
         """
-        if dbfolder == "\\":
-            return
-        if not dbfolder.endswith("\\"):
-            dbfolder += "\\"
-        self.connection = sqlite3.connect(f'{dbfolder}orchestrator.db')
+        if os.name == 'nt':
+            if dbfolder == "\\":
+                return
+            if not dbfolder.endswith("\\"):
+                dbfolder += "\\"
+        else:
+            if not dbfolder.endswith("/"):
+                dbfolder += "/"
+        self.connection = connect(f'{dbfolder}orchestrator.db')
         self.connection.execute("PRAGMA foreign_keys = 1")
         self.connection.execute("PRAGMA JOURNAL_MODE = 'WAL'")
+        # if os.name == 'posix':
+        #    self.connection.execute("PRAGMA SQLITE_ENABLE_LOCKING_STYLE = 1")
+        #    self.connection.execute("PRAGMA journal=OFF")
+        self.error = None
 
     def run_sql(self, sql, tablename: str = ""):
         """
@@ -1174,11 +1256,33 @@ class SQL:
         if len(tablename) > 0:
             try:
                 return self.get_inserted_id(tablename)
-            except (ValueError, Exception):
-                return None
+            except Exception as ex:
+                self.set_error(ex)
+                raise Exception(self.error)
         else:
             return None
 
+    def set_error(self, ex: any):
+        """
+        Set the internal error comming from the try-except
+        :param ex: the exception
+        """
+        trace = []
+        tb = ex.__traceback__
+        while tb is not None:
+            trace.append({
+                "filename": tb.tb_frame.f_code.co_filename,
+                "name": tb.tb_frame.f_code.co_name,
+                "lineno": tb.tb_lineno
+            })
+            tb = tb.tb_next
+            err_ = str({
+                'type': type(ex).__name__,
+                'message': str(ex),
+                'trace': trace
+            })
+        self.error = err_
+
     def get_inserted_id(self, tablename: str) -> int:
         """
         Get the last inserted id of the primary key column of the table
@@ -1189,13 +1293,19 @@ class SQL:
         curs = self.connection.cursor()
         curs.execute(sql)
         row = curs.fetchone()
+        if row[0] is None:
+            return None
         return int(row[0])
 
     def commit(self):
         """
         Commit any sql statement
         """
-        self.connection.commit()
+        try:
+            self.connection.commit()
+        except Exception as ex:
+            self.set_error(ex)
+            raise Exception(self.error)
 
     def get_saved_flows(self):
         """
@@ -1243,7 +1353,7 @@ class SQL:
             ret.append([rw[0], rw[1], rw[2], rw[3], rw[4]])
         return ret
 
-    def remove_saved_flows(self, lst: List = None):
+    def remove_saved_flows(self, lst: list = None):
         """
         Removes saved flows from the orchestrator database by matching on the given list of flow-names
         :param lst: The list with flow names to remove from the database
@@ -1260,14 +1370,18 @@ class SQL:
         """
         Create tables for the Orchestrator database
         """
-        sql = "CREATE TABLE IF NOT EXISTS Flows (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, name TEXT NOT NULL, location TEXT NOT NULL, description TEXT, timestamp DATE DEFAULT (datetime('now','localtime')));"
-        self.run_sql(sql)
-        sql = "CREATE TABLE IF NOT EXISTS Runs (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, flow_id INTEGER NOT NULL, name TEXT NOT NULL, result TEXT, started DATE DEFAULT (datetime('now','localtime')), finished DATE DEFAULT (datetime('now','localtime')), CONSTRAINT fk_saved FOREIGN KEY (flow_id) REFERENCES Flows (id) ON DELETE CASCADE);"
-        self.run_sql(sql)
-        sql = "CREATE TABLE IF NOT EXISTS Steps (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, run INTEGER NOT NULL, status TEXT, name TEXT NOT NULL,step TEXT,result TEXT,timestamp DATE DEFAULT (datetime('now','localtime')), CONSTRAINT fk_runs FOREIGN KEY (run) REFERENCES Runs (id) ON DELETE CASCADE);"
-        self.run_sql(sql)
-        # sql = "CREATE TABLE IF NOT EXISTS Triggers (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, flow_id INTEGER NOT NULL, trigger_info, CONSTRAINT flows_saved_trigger FOREIGN KEY (flow_id) REFERENCES Flows (id) ON DELETE CASCADE);"
-        # self.run_sql(sql)
+        try:
+            sql = "CREATE TABLE IF NOT EXISTS Flows (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, name TEXT NOT NULL, location TEXT NOT NULL, description TEXT, timestamp DATE DEFAULT (datetime('now','localtime')));"
+            self.run_sql(sql)
+            sql = "CREATE TABLE IF NOT EXISTS Runs (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, flow_id INTEGER NOT NULL, name TEXT NOT NULL, result TEXT, started DATE DEFAULT (datetime('now','localtime')), finished DATE DEFAULT (datetime('now','localtime')), CONSTRAINT fk_saved FOREIGN KEY (flow_id) REFERENCES Flows (id) ON DELETE CASCADE);"
+            self.run_sql(sql)
+            sql = "CREATE TABLE IF NOT EXISTS Steps (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, run INTEGER NOT NULL, status TEXT, name TEXT NOT NULL,step TEXT,result TEXT,timestamp DATE DEFAULT (datetime('now','localtime')), CONSTRAINT fk_runs FOREIGN KEY (run) REFERENCES Runs (id) ON DELETE CASCADE);"
+            self.run_sql(sql)
+            # sql = "CREATE TABLE IF NOT EXISTS Triggers (id INTEGER NOT NULL PRIMARY KEY AUTOINCREMENT, flow_id INTEGER NOT NULL, trigger_info, CONSTRAINT flows_saved_trigger FOREIGN KEY (flow_id) REFERENCES Flows (id) ON DELETE CASCADE);"
+            # self.run_sql(sql)
+        except Exception as ex:
+            self.set_error(ex)
+            raise Exception(self.error)
 
 
 class Visio:
@@ -1386,7 +1500,7 @@ class Visio:
                 retn.append(step)
         return retn
 
-    def get_step_from_shape(self, shape: Any) -> Any:
+    def get_step_from_shape(self, shape: any) -> any:
         """
         Build a Step-object from the Shape-object
         :param shape: The Shape-object
@@ -1527,6 +1641,6 @@ class Visio:
 
 # Test
 # engine = WorkflowEngine()
-# doc = engine.open(fr"c:\\temp\\test.vsdx")  # c:\\temp\\test.xml
+# doc = engine.open(fr"c:\\temp\\test2.flw")  # c:\\temp\\test.xml
 # steps = engine.get_flow(doc)
 # engine.run_flow(steps)
